
mkdir build && cd build

cmake -LAH -G"NMake Makefiles" ^
  -DCMAKE_BUILD_TYPE=Release ^
  -DCMAKE_PREFIX_PATH=%LIBRARY_PREFIX% ^
  -DCMAKE_INSTALL_PREFIX=%LIBRARY_PREFIX% ^
  -DBUILD_SHARED_LIBS=ON ^
  -DINSTALL_LIBS=ON ^
  -DBUILD_PYBULLET=ON ^
  -DBUILD_PYBULLET_NUMPY=ON ^
  -DBUILD_CPU_DEMOS=ON ^
  -DBUILD_UNIT_TESTS=OFF ^
  -DBUILD_EXTRAS=ON ^
  -DBUILD_OPENGL3_DEMOS=ON ^
  -DPYTHON_EXECUTABLE=%PYTHON% ^
  -DPYTHON_LIBRARY=%PREFIX%/libs/python%CONDA_PY%.lib ^
  -DPYTHON_VERSION_PYBULLET=%PY_VER% ^
  -DPYTHON_SITE_PACKAGES=%SP_DIR% ^
  ..
if errorlevel 1 exit 1

cmake --build . --config %CMAKE_CONFIG% --target install
if errorlevel 1 exit 1
